
;; ===========================================================================
;; testing ODBC functionality
;; ===========================================================================
;; yet compatible to gold hill and clisp
;; ===========================================================================

(defvar *henv* nil)
(defvar *hdbc* nil)
(defvar *hstmt* nil)
(defvar *errors* nil)
;; gclisp has a problem at gc time
(defvar row nil)

(defun odbc_success (ret-val)
  (or (eql ret-val SQL_SUCCESS) (eql ret-val SQL_SUCCESS_WITH_INFO)))

(defun append_error (text)
  (setf *errors* (append *errors* (list text))))

(defun clear_errors ()
  (setf *errors* nil))

(defun no_errors ()
  (zerop (length *errors*)))

(defun ce ()
  (clear_errors))

(defun commit ()
  (let (ret)
    (when *hdbc*
      (setf ret (SQLEndTran SQL_HANDLE_DBC (%VAL *hdbc*) SQL_COMMIT)))
  (when (odbc_success ret) "OK")))

(defun connect (dsn uid pwd)
  (let ((ret nil))
    (setf *henv*  (make-handle))
    (setf *hdbc*  (make-handle))
    (setf *hstmt* (make-handle))
    (setq ret (SQLAllocHandle SQL_HANDLE_ENV SQL_NULL_HANDLE (%ADR *henv*)))
    (unless (odbc_success ret)
      (append_error "failed to allocate environment handle")
      (return-from connect))
    (setf ret (SQLSetEnvAttr (%VAL *henv*) SQL_ATTR_ODBC_VERSION SQL_OV_ODBC3 SQL_NTS))
    (unless (odbc_success ret)
      (SQLFreeHandle SQL_HANDLE_ENV (%VAL *henv*))
      (append_error "setting environment to ODBC 3 failed")
      (return-from connect))
    (setq ret (SQLAllocHandle SQL_HANDLE_DBC (%VAL *henv*) (%ADR *hdbc*)))
    (unless (odbc_success ret)
      (SQLFreeHandle SQL_HANDLE_ENV (%VAL *henv*))
      (append_error "allocation of connection handle failed")
      (return-from connect))
    ;(setq ret (SQLConnect (%VAL *hdbc*) "LispTest" "admin" ""))
    ;(setq ret (SQLConnect (%VAL *hdbc*) "OraTest" "system" "manager"))
    (setq ret (SQLConnect (%VAL *hdbc*) dsn uid pwd))
    (unless (odbc_success ret)
      (SQLFreeHandle SQL_HANDLE_ENV (%VAL *henv*))
      (SQLFreeHandle SQL_HANDLE_DBC (%VAL *hdbc*))
      (append_error "connection could not be established")
      (return-from connect))
    )
  "OK")

(defun disconnect ()
  (let ((ret nil))
    (unless *hdbc*
      (append_error "not connected")
      (return-from disconnect))
    (setq ret (SQLDisconnect (%VAL *hdbc*)))
    (unless (odbc_success ret)
      (append_error "disconnect from database failed"))
    (setq ret (SQLFreeHandle SQL_HANDLE_DBC (%VAL *hdbc*)))
    (setf *hdbc* nil)
    (unless (odbc_success ret)
      (append_error "release of db handle failed"))
    (setq ret (SQLFreeHandle SQL_HANDLE_ENV (%VAL *henv*)))
    (unless (odbc_success ret)
      (format t "release of environment handle failed!~%"))
  )
  (when (no_errors) "OK"))

(defun sql-exec-i (statement)
  (let (ret)
    (setq ret (SQLAllocHandle SQL_HANDLE_STMT (%VAL *hdbc*) (%ADR *hstmt*)))
    (unless (odbc_success ret)
      (append_error "allocation of statement handle failed")
      (return-from sql-exec-i))
    (setq ret (SQLExecDirect (%VAL *hstmt*) statement))
    (unless (odbc_success ret)
      (append_error "execution of statement failed"))
    (sql-diag)
  )
  (when (no_errors) "OK"))

(defun sql-exec (statement)
  (let (ret)
    (sql-exec-i statement)
    (sql-diag)
    (setq ret (SQLFreeHandle SQL_HANDLE_STMT (%VAL *hstmt*)))
    (unless (odbc_success ret)
      (append_error "release of statement handle failed")
      (return-from sql-exec))
  )
  (when (no_errors) "OK"))

(defun sql-diag ()
  (let ((sqlstate (make-c-string 8))
        (sqlnative (make-c-long))
        (sqlmsg  (make-c-string 256))
        (txtlen (make-c-long))
        ret)
    (setf ret (SQLGetDiagRec SQL_HANDLE_STMT (%val *hstmt*) 1 
                             (%adr sqlstate) (%adr sqlnative) (%adr sqlmsg) 256 (%adr txtlen)))
    (when (odbc_success ret)
      (format t "~%~a" (%val sqlmsg)))
    ))

(defun describe-col (hstmt index)
  (let ((ret nil)
        (ColumnNamePtr (make-c-string 80))
        (NameLengthPtr      (make-c-long))
        (DataTypePtr        (make-c-int))
        (ColumnSizePtr      (make-c-long))
        (DecimalDigitsPtr   (make-c-long))
        (NullablePtr        (make-c-long)))
    (setf ret (SQLDescribeCol (%VAL hstmt)
                              index
                              (%ADR ColumnNamePtr)
                              80
                              (%ADR NameLengthPtr)
                              (%ADR DataTypePtr)
                              (%ADR ColumnSizePtr)
                              (%ADR DecimalDigitsPtr)
                              (%ADR NullablePtr)))
    (if (not (odbc_success ret))
      (progn
        (append_error (format nil "SQLDescribeCol failed with: ~a" ret))
        (setf ret nil))
      (setf ret 
        (list :name           (%VAL ColumnNamePtr)
              :name-length    (%VAL NameLengthPtr)
              :data-type      (%VAL DataTypePtr)
              :column-size    (%VAL ColumnSizePtr)
              :decimal-digits (%VAL DecimalDigitsPtr)
              :nullable       (%VAL NullablePtr))))))

(defun dsc-col (index)
  (describe-col *hstmt* index))

(defun query (statement)
  (let ((col-count-ptr (make-c-int))
        (ret nil)
        (cols nil)
        (size nil)
        (row nil))
    (clear_errors)
    (sql-exec-i statement)
    (when (no_errors)
      (setf ret (SQLNumResultCols (%VAL *hstmt*) (%ADR col-count-ptr)))
      (unless (odbc_success ret)
        (append_error (format t "NumResultCols returned ~a" ret))
        (return-from query))
      (setf cols (%VAL col-count-ptr))
      (setf row nil)
      (let ((i 0)
            (tmp1)
            (line ""))
        (setf row nil)
        (loop
          (when (> (incf i) cols) (return))
          (print (dsc-col i))
          (setf line (string-append line (format nil "~12@a" (getf (dsc-col i) :name))))
          (setf tmp1 (make-c-long))
          (setf (%VAL tmp1) 0)
          (setf size (getf (dsc-col i) :column-size))
          (when (> size 4096) (setf size 4096))
          (setf row (append row (list (make-c-string size) tmp1)))
          (setf ret (SQLBindCol (%VAL *hstmt*)
                      i
                      SQL_C_CHAR
                      (%ADR (nth (* (1- i) 2) row))
                      255
                      (%ADR (nth (1+ (* (1- i) 2)) row))))
          (unless (odbc_success ret) 
            (append_error (format nil "SQLBindCol returned: ~a" ret))
            (return-from query))
          )
        (unless (zerop (length line))
          (print line)
          (print (make-string 79 :initial-element #\-)))
        )
      (unless (zerop cols)
        (loop
          (setf ret (SQLFetch (%VAL *hstmt*)))
          (unless (odbc_success ret)
            ;(print (format nil "SQLFetch returned: ~a" ret))
            (return))
          (let ((i 0)
                (tmp nil)
                (line ""))
            (loop
              (when (> (incf i) cols) (return))
              (setf tmp (nth (* (1- i) 2) row))
              (when tmp (setf tmp (%VAL tmp)))
              (setf tmp (string-trim '(#\Space) tmp))
              (setf tmp1 (%VAL (nth (1+ (* (1- i) 2)) row)))
              (when (= tmp1 -1) (setf tmp "NULL"))
              (setf line (string-append line (format nil "~12@a" tmp)))
              ;(setf line (string-append line (format nil "~12@a" tmp1)))
              )
            (print line))))
      (setq ret (SQLFreeHandle SQL_HANDLE_STMT (%VAL *hstmt*)))
      (unless (odbc_success ret)
        (append_error "release of statement handle failed"))
      ))
  (if (no_errors) 
    "OK"
    *errors*))

(defun sql ()
  (format t "~%")
  (let ((line)
        (ret))
    (loop
      (princ "SQL>")
      (setf line (read-line *standard-input* nil nil))
      ;(setf line (read-line))
      ;(print line)
      (when line
        (setf line (string-trim '(#\Space #\Tab) line)))
      (ce)
      (cond ((equalp line "exit") (return))
            ((equalp line "commit") (setf ret (commit)) (setf line nil))
            ;((equalp line "connect") (setf ret (connect)) (setf line nil))
            ((equalp line "disconnect") (setf ret (disconnect)) (setf line nil))
            ((equalp line "exit") (disconnect) (exit))
            )
      (when line
        (ce)
        (setf ret (query line)))
        (format t "~%~a~%" ret)
      ;(format t "~%room:~a~%" (room))
      (format t "~%")
      )
    )
  )

(defun job ()
  (let ((i 1)
        (basename "test")
        (table "")
        (ret))
    (clear_errors)
    (connect)
    (loop
      (setf table (format nil "T_TEST_~a" i))
      (when (execute (format nil "create table ~a (a char(10), b number)" table))
        (return))
      (clear_errors)
      (when (> (incf i) 20) 
        (append_error "table limit reached")
        (return)))
    (setf i 1)
    (loop
      (execute (format nil "insert into ~a values (~a, ~a)" table i (* i 2)))
      (when (> (incf i) 10) (return)))
    (disconnect)
    (when (no_errors) (format nil "table ~a created and filled" table))))

;(sql)







